import json
import httpx


class APIError(Exception):
    """基础API异常类"""

    def __init__(self, message: str, code: int | None = None):
        self.code = code
        self.message = message
        super().__init__(f"[{code}] {message}" if code else message)


class AsyncChatClient:
    def __init__(
        self,
        endpoint: str,
        api_key: str,
        timeout: int = 120,
    ):
        """
        初始化异步聊天客户端

        :param endpoint: API端点
        :param api_key: API密钥
        :param timeout: 请求超时时间（秒）
        """
        self.endpoint = endpoint
        self.api_key = api_key
        self.base_url = endpoint
        self.timeout = timeout
        self._client = httpx.AsyncClient()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc_info):
        await self.close()

    async def close(self):
        """关闭客户端"""
        await self._client.aclose()

    def _build_headers(self) -> dict[str, str]:
        """构建请求头"""
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        return headers

    async def create(self, model: str, messages: list, stream: bool = False, **kwargs):
        """
        创建聊天请求

        :param model: 使用的模型名称
        :param messages: 消息列表
        :param stream: 是否使用流式响应
        :return: 生成器（流式模式）或字典（非流式）
        """
        url = f"{self.base_url}/chat/completions"

        # 构建基础请求参数
        payload = {"model": model, "messages": messages, "stream": stream, **kwargs}

        # 发送请求
        response = await self._client.post(
            url, headers=self._build_headers(), json=payload, timeout=self.timeout
        )

        if response.status_code != 200:
            try:
                error_data = response.json()
                print(error_data)
                raise APIError(
                    error_data.get("message", "Unknown error"),
                    code=error_data.get("code", response.status_code),
                )
            except json.JSONDecodeError:
                raise APIError(
                    f"HTTP Error {response.status_code}", code=response.status_code
                )
        return response

    async def stream_response(self, response: httpx.Response):
        self.content = ""
        async for chunk in response.aiter_lines():
            if not chunk.strip() or chunk.startswith(": ping"):
                continue
            try:
                if chunk.startswith("data: "):
                    data: dict = json.loads(chunk[6:])
                    if data["finish_reason"] == "stop":
                        self.content += data["choices"][0]["delta"]["content"]
                    yield data
            except json.JSONDecodeError:
                continue

    async def non_stream_response(self, response: httpx.Response):
        return response.json()




from pydantic import BaseModel, field_validator


class Config(BaseModel):
    zssm_ai_text_endpoint: str = "https://api.deepseek.com/v1"
    zssm_ai_text_token: str | None = None
    zssm_ai_text_model: str = "deepseek-chat"
    
    zssm_ai_vl_endpoint: str = "https://api.siliconflow.cn/v1"
    zssm_ai_vl_token: str | None = None
    zssm_ai_vl_model: str = "Qwen/Qwen2.5-VL-72B-Instruct"
    
    zssm_browser_proxy: str | None = None

    @field_validator("zssm_ai_text_endpoint")
    def check_zssm_ai_text_endpoint(cls, v):
        if not v.startswith("http://") and not v.startswith("https://"):
            raise ValueError("zssm_ai_text_endpoint must start with http:// or https://")
        return v

    @field_validator("zssm_ai_text_token")
    def check_zssm_ai_text_token(cls, v):
        if not v:
            raise ValueError("zssm_ai_text_token must not be empty")
        return v

    @field_validator("zssm_ai_text_model")
    def check_zssm_ai_text_model(cls, v):
        if not v:
            raise ValueError("zssm_ai_text_model must not be empty")
        return v



from loguru import logger
from playwright.async_api import Browser, Error, Playwright, async_playwright

from .installer import install_browser

_browser: Browser | None = None
_playwright: Playwright | None = None


async def init(**kwargs) -> Browser:
    global _browser  # noqa: PLW0603
    global _playwright  # noqa: PLW0603
    _playwright = await async_playwright().start()
    try:
        _browser = await launch_browser(**kwargs)
    except Error:
        await install_browser()
        _browser = await launch_browser(**kwargs)
    return _browser


async def launch_browser(**kwargs) -> Browser:
    assert _playwright is not None, "Playwright 没有安装"
    logger.info("使用 firefox 启动")
    return await _playwright.firefox.launch(**kwargs)


async def get_browser(**kwargs) -> Browser:
    return _browser if _browser and _browser.is_connected() else await init(**kwargs)



import asyncio
import os
import re
import sys
from pathlib import Path

from loguru import logger
from playwright._impl._driver import compute_driver_executable, get_driver_env


def log(level: str, rich_text: str) -> None:
    getattr(logger.opt(colors=True), level)(
        rich_text.replace("[", "<").replace("]", ">"),
        alt=rich_text,
    )


class Progress:
    def __init__(self, name: str) -> None:
        self.last_updated: float = 0
        self.progress: float = 0.0
        self.name = name

    def update(self, *, target: float):
        import time

        if self.progress >= 100:
            return
        if (
            time.time() - self.last_updated >= 1
            or (target - self.progress >= 10 and time.time() - self.last_updated >= 0.1)
            or target == 100
        ):
            self.progress = target
            log(
                "info",
                f"[cyan]{self.name}[/] [green]{'-' * int(self.progress / 5):<20}[/] [magenta]{int(self.progress)}%[/]",
            )
            self.last_updated = time.time()


async def install_browser(
    download_host: str | None = None,
    browser_type: str = "firefox",
):
    env = get_driver_env()
    if download_host:
        env["PLAYWRIGHT_DOWNLOAD_HOST"] = download_host

    cde_raw = compute_driver_executable()
    cde = [cde_raw.as_posix()] if isinstance(cde_raw, Path) else list(cde_raw)

    command = [
        *cde,
        "install",
        "--with-deps",
        browser_type,
    ]

    if sys.platform.startswith("win") or os.name == "nt":
        log(
            "info",
            f"Start download Playwright for {browser_type} with dependencies, "
            "may require administrator privileges from you."
            f"command: [cyan]{' '.join(command)}[/]",
        )
    else:
        log(
            "info",
            f"Start download Playwright for {browser_type} with dependencies, may require you to access sudo.",
        )

    shell = await asyncio.create_subprocess_exec(*command, stdout=asyncio.subprocess.PIPE, env=env)
    returncode = None

    assert shell.stdout

    progress: Progress | None = None

    while line := re.sub("\x1b.*?m", "", (await shell.stdout.readline()).decode("UTF-8")):
        if "Downloading" in line:
            progress = Progress(line[12:-1])
        if percent := re.findall("(\\d+)%", line):
            progress_target = float(percent[0])
            if progress:
                progress.update(target=progress_target)
        elif p := re.match("(?P<file>.*) downloaded to (?P<path>.*)", line):
            p = p.groupdict()
            log(
                "success",
                "Downloaded [cyan]{file}[/] to [magenta]{path}[/]".format(file=p["file"], path=p["path"]),
            )
        elif line == "Failed to install browsers\n":
            message = await shell.stdout.read()
            log("error", "Download Failed:\n" + message.decode("UTF-8"))
            returncode = 1

    if returncode or shell.returncode:
        log("error", f"Failed to download Playwright for {browser_type}.")
        log("error", "Please see: [magenta]https://playwright.dev/python/docs/intro[/]")
        log(
            "error",
            "Run [magenta]poetry run playwright install[/] or "
            "[magenta]pdm run playwright install[/] to install Playwright manually.",
        )
    else:
        log("success", f"Playwright for {browser_type} is installed.")


from nonebot import require, get_driver
from nonebot.plugin import PluginMetadata

from .browser import install_browser


require("nonebot_plugin_alconna")

__plugin_meta__ = PluginMetadata(
    name="nonebot-plugin-zssm",
    description="这是什么？问一下！用 ai 来解释群友发送的「未知」内容",
    usage="对着任意你不懂的内容发送「zssm」即可",
    homepage="https://github.com/djkcyl/nonebot-plugin-zssm",
    type="application",
    supported_adapters={"~onebot.v11", "~onebot.v12", "~qq"},
    extra={"author": "djkcyl", "version": "0.1.8"},
)

from . import handle  # noqa

driver = get_driver()
driver.on_startup(install_browser)



from nonebot_plugin_alconna import on_alconna
from nonebot import get_plugin_config
from nonebot_plugin_alconna.builtins.extensions.reply import ReplyRecordExtension
from nonebot.internal.adapter import Event, Bot
from nonebot_plugin_alconna.uniseg import Reply, UniMessage, Text, MsgId, Image
from nonebot_plugin_alconna.builtins.uniseg.market_face import MarketFace
from .api import AsyncChatClient
from contextlib import suppress
from .config import Config
from pathlib import Path
from PIL import Image as PILImage
from io import BytesIO
from .browser import get_browser
import httpx
import base64
from loguru import logger
import random
import re
import ssl

system_prompt_raw = (
    Path(__file__).parent.joinpath("prompt.txt").read_text(encoding="utf-8")
)
config = get_plugin_config(Config)

zssm = on_alconna("zssm", extensions=[ReplyRecordExtension()])


async def url_to_base64(url: str) -> str:
    ssl_context = ssl.create_default_context()
    ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_TLSv1_3
    ssl_context.set_ciphers("HIGH:!aNULL:!MD5")
    async with httpx.AsyncClient(verify=ssl_context) as client:
        response = await client.get(url)
        image = PILImage.open(BytesIO(response.content))
        # 把图片控制在5mb以内
        if len(response.content) > 5 * 1024 * 1024:
            image.thumbnail((4096, 4096))
        if image.mode != "RGB":
            image = image.convert("RGB")
        buffered = BytesIO()
        image.save(buffered, format="JPEG", quality=80)
        b64_content = base64.b64encode(buffered.getvalue()).decode()
        return f"data:image/jpeg;base64,{b64_content}"


@zssm.handle()
async def handle(msg_id: MsgId, ext: ReplyRecordExtension, event: Event, bot: Bot):
    msg = event.get_message()
    if reply := ext.get_reply(msg_id):
        reply_msg_raw = reply.msg
    else:
        return await UniMessage(Text("未找到上一条消息")).send(reply_to=Reply(msg_id))

    if not reply.msg:
        return await UniMessage(Text("上一条消息内容为空")).send(reply_to=Reply(msg_id))

    if not config.zssm_ai_text_token or not config.zssm_ai_vl_token:
        return await UniMessage(Text("未配置 Api Key，暂时无法使用")).send(
            reply_to=Reply(msg_id)
        )

    with suppress(Exception):
        await bot.call_api("set_msg_emoji_like", message_id=msg_id, emoji_id=424)

    if isinstance(reply_msg_raw, str):
        reply_msg_raw = msg.__class__(reply_msg_raw)

    reply_msg = UniMessage.generate_sync(message=reply_msg_raw)
    image_list = reply_msg.get(Image)
    reply_msg_display = ""
    for item in reply_msg:
        if isinstance(item, Image):
            reply_msg_display += f"[图片 {item.id}]"
        if isinstance(item, MarketFace):
            return await UniMessage(Text("不支持商城表情")).send(reply_to=Reply(msg_id))
        else:
            reply_msg_display += str(item)

    random_number = str(random.randint(10000000, 99999999))
    system_prompt = system_prompt_raw + random_number
    user_prompt = f"<random number: {random_number}> \n<type: text>\n{reply_msg_display}\n</type: text>\n"

    for image in image_list:
        image: Image
        image_url = image.url
        if not image_url:
            continue
        async with AsyncChatClient(
            config.zssm_ai_vl_endpoint, config.zssm_ai_vl_token
        ) as client:
            logger.info(f"image_url: {image_url}")
            try:
                response = await client.create(
                    config.zssm_ai_vl_model,
                    [
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": await url_to_base64(image_url)
                                    },
                                },
                                {"type": "text", "text": "请描述这张图片"},
                            ],
                        },
                    ],
                )
            except Exception as e:
                logger.error(e)
                return await UniMessage(Text("图片识别失败")).send(
                    reply_to=Reply(msg_id)
                )
            image_prompt = response.json()["choices"][0]["message"]["content"]
            user_prompt += (
                f"<type: image, id: {image.id}>\n{image_prompt}\n</type: image>\n"
            )

    reg_match = re.compile(
        r"\b(?:https?|ftp):\/\/[^\s\/?#]+[^\s]*|\b(?:www\.)?[a-zA-Z0-9-]+\.[a-zA-Z]{2,}(?:\.[a-zA-Z]{2,})*(?:\/[^\s]*)?\b"
    )
    msg_url_list = reg_match.findall(str(reply_msg))
    if msg_url_list:
        browser = await get_browser(
            proxy={"server": config.zssm_browser_proxy}
            if config.zssm_browser_proxy
            else None
        )
        msg_url = msg_url_list[0]
        logger.info(f"msg_url: {msg_url}")
        await UniMessage(Text("正在尝试打开消息中的第一条链接")).send(
            reply_to=Reply(msg_id)
        )
        page = await browser.new_page()
        try:
            await page.goto(msg_url, timeout=60000)
        except Exception:
            return await UniMessage(Text("打开链接失败")).send(reply_to=Reply(msg_id))
        # 获取页面的内容
        page_content = await page.query_selector("html")
        if page_content:
            page_content = await page_content.inner_text()
            user_prompt += (
                f"<type: web_page, url: {msg_url}>\n{page_content}\n</type: web_page>\n"
            )
        await page.close()
        if not page_content:
            return await UniMessage(Text("无法获取页面内容")).send(
                reply_to=Reply(msg_id)
            )

    if msg_url_list or image_list:
        with suppress(Exception):
            await bot.call_api("set_msg_emoji_like", message_id=msg_id, emoji_id=314)

    user_prompt += f"</random number: {random_number}>\n"
    logger.info(f"user_prompt: \n{user_prompt}")
    async with AsyncChatClient(
        config.zssm_ai_text_endpoint, config.zssm_ai_text_token
    ) as client:
        response = await client.create(
            config.zssm_ai_text_model,
            [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_prompt},
            ],
        )
        logger.info(response.json())

    await UniMessage(Text(response.json()["choices"][0]["message"]["content"])).send(
        reply_to=Reply(msg_id)
    )
